from flask import Blueprint, render_template, request
import pandas as pd
import plotly.express as px
import json
import plotly.utils


chairman_blueprint = Blueprint('chairman', __name__, template_folder="../templates")


# FUNCTION TO GET CSV DATA
def getCsvData(filename):
    # Load CSV into a DataFrame
    data = pd.read_csv(filename)
    # Get column names
    headers = data.columns.tolist()
    # Convert the DataFrame to a list of dictionaries for rendering in HTML
    table_data = data.to_dict(orient="records")
    
    return headers, table_data



# FUNCTION TO CLASSIFY DATA
def classifyData(table_data):
    # Convert table_data (list of dicts) to a DataFrame
    df = pd.DataFrame(table_data)
    
    # Ensure SDI is numeric (convert errors to NaN)
    df['SDI'] = pd.to_numeric(df['SDI'], errors='coerce')
    
    # Initial grouping by 'Initiative'
    grouped = df.groupby('Initiative').size().reset_index(name='Project Count')
    
    # Create a new DataFrame for delays, filtering out certain Status_Final values
    delay_df = df[~df['Status_Final'].isin(['On Track', 'On Hold', 'Finished', 'Closure', 'Not Started'])]
    
    # Calculate delayed project counts using delay_df
    delay_counts = delay_df.groupby('Initiative').apply(
        lambda g: pd.Series({
            'Delayed Projects': (g['SDI'] > 0).sum()
        })
    ).reset_index()

    # Calculate delayed projects by each executor within each initiative
    executor_counts = delay_df.groupby(['Initiative', 'Executer']).size().reset_index(name='Executor Delayed Projects')

    # Create a wide format table for each executor (with executors as columns)
    executor_pivot = executor_counts.pivot_table(
        index='Initiative', 
        columns='Executer', 
        values='Executor Delayed Projects', 
        aggfunc='sum', 
        fill_value=0
    )

    # Merge grouped delay counts with the executor counts
    summary_data = grouped.merge(delay_counts, on='Initiative', how='left').fillna(0)

    # Merge executor counts (dynamic columns) with the summary_data
    summary_data = summary_data.set_index('Initiative').join(executor_pivot).reset_index()

    # Add a total column for all executor delays
    executor_columns = [col for col in executor_pivot.columns]
    summary_data['Total Executer Delayed Projects'] = summary_data[executor_columns].sum(axis=1)

    # Calculate total project counts for each executor across all initiatives
    total_projects_by_executor = df.groupby('Executer').size().to_dict()

    # Include executor project totals in summary_data for template rendering
    summary_data_meta = {
        'executor_columns': executor_columns,
        'total_projects_by_executor': total_projects_by_executor
    }

    # Convert to a list of dictionaries for rendering in HTML
    summary_data = summary_data.to_dict(orient="records")
    return summary_data, summary_data_meta, delay_df

@chairman_blueprint.route("/chairman/chart")
def generate_chart():
    try:
        initiative = request.args.get("initiative")
        if not initiative:
            return {"error": "No initiative specified"}, 400

        # Get CSV data
        _, table_data = getCsvData("V3_Data.csv")

        # Convert to DataFrame
        df = pd.DataFrame(table_data)

        # Ensure SDI and Delay_Percent are numeric (convert errors to NaN)
        df['SDI'] = pd.to_numeric(df['SDI'], errors='coerce')
        df['Delay_Percent'] = pd.to_numeric(df['Delay_Percent'], errors='coerce')

        # Drop rows where Status_Final is in the specified list
        excluded_statuses = ["On Track", "On Hold", "Finished", "Closure", "Not Started"]
        df = df[~df["Status_Final"].isin(excluded_statuses)]

        # Filter by the selected initiative and SDI > 0
        filtered_df = df[(df["Initiative"] == initiative) & (df["SDI"] > 0)]

        if filtered_df.empty:
            return {"error": "No valid data available for this initiative"}, 404

        # Ensure the SN column exists and is used for unique project identification
        if "SN" not in filtered_df.columns:
            return {"error": "SN column missing from data"}, 500

        # Ensure the Executer column exists
        if "Executer" not in filtered_df.columns:
            return {"error": "Executer column missing from data"}, 500

        # Create Unique Project Name using Executor + (SN)
        filtered_df['Unique_Project_Name'] = (
            filtered_df['Executer'] + ' (' + filtered_df['SN'].astype(str) + ')'
        )

        # Melt the DataFrame with Delay_Percent first, then SDI
        melted_df = filtered_df.melt(
            id_vars=["Unique_Project_Name"],
            value_vars=["Delay_Percent", "SDI"],  # Delay_Percent first
            var_name="Metric", value_name="Value"
        )

        # Map Metric names to desired legend labels
        metric_mapping = {
            'Delay_Percent': 'Progress Delay - Current',
            'SDI': 'Performance Delay - SDI'
        }
        melted_df['Metric'] = melted_df['Metric'].map(metric_mapping)

        # Generate the chart
        fig = px.bar(
            melted_df,
            x="Unique_Project_Name",
            y="Value",
            color="Metric",
            text="Value",  # Add value labels on top of bars
            barmode="group",  # Side-by-side bars
            title=f"Delayed Projects in Initiative: {initiative}",  # Initial title setting
            labels={
                "Unique_Project_Name": "Executor (SN)",
                "Value": "Metric Value",
                "Metric": "Type"
            },
        )

        # Customize traces for colors and patterns based on new Metric names
        for trace in fig.data:
            if trace.name == "Progress Delay - Current":
                trace.marker.color = "red"  # Solid red color for Progress Delay - Current
            elif trace.name == "Performance Delay - SDI":
                trace.marker.color = "lightgrey"  # Grey color for Performance Delay - SDI
                trace.marker.pattern = dict(shape="x")  # Hatched pattern for Performance Delay - SDI

        # Add the red line at y=10
        fig.add_shape(
            type="line",
            x0=-0.5,  # Extend the line slightly left of the first bar
            x1=len(filtered_df["Unique_Project_Name"]) - 0.5,  # Extend to the last bar
            y0=10,
            y1=10,
            line=dict(color="red", width=1, dash="dash"),
            xref="x",
            yref="y"
        )

        # Add border to the plot area
        fig.add_shape(
            type="rect",
            x0=0, x1=1, y0=0, y1=1,  # Full plot area
            xref="paper",  # Relative to the paper size
            yref="paper",
            line=dict(color="black", width=1)  # Solid black border
        )

        # Update y-axis configuration with gridlines
        fig.update_yaxes(
            type="log",  # Logarithmic scale
            range=[0, 2.48],  # Updated range to cover up to 300 (10^2.48 ≈ 300)
            tickvals=[1, 2, 4, 6, 8, 10, 25, 50, 75, 100, 150, 200, 250, 300],  # Added 300
            title="Delay %",
            automargin=True,  # Automatically adjust margins
            ticklabelposition="outside",  # Ensure labels are placed outside
            showgrid=True,  # Add horizontal gridlines
            gridcolor="lightgrey",  # Gridline color
            gridwidth=1  # Gridline width
        )

        # Update x-axis configuration with gridlines
        fig.update_xaxes(
            tickangle=90,  # Vertical labels
            title="Executor (SN)",
            automargin=True,  # Automatically adjust margins
            ticklabelposition="outside",  # Ensure labels are placed outside
            showgrid=True,  # Add vertical gridlines
            gridcolor="lightgrey",  # Gridline color
            gridwidth=1  # Gridline width
        )

        # Adjust overall layout to add padding, reposition legend, and center title
        fig.update_layout(
            plot_bgcolor="rgba(0,0,0,0)",  # Transparent background
            paper_bgcolor="rgba(255,255,255,1)",  # White border
            margin=dict(l=60, r=60, t=120, b=60),  # Adjust overall plot margins (increased top margin for legend)
            legend=dict(
                orientation="h",  # Horizontal orientation
                yanchor="bottom",
                y=1.02,  # Position above the plot
                xanchor="center",
                x=0.5  # Centered horizontally
            ),
            title=dict(
                text=f"Delayed Projects in Initiative: {initiative}",  # Chart title
                x=0.5,  # Center horizontally
                y=0.95,  # Position vertically (adjust as needed)
                xanchor='center',
                yanchor='top',
                font=dict(
                    size=20  # Adjust font size as needed
                )
            )
        )

        # Convert the chart to JSON
        chart_json = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return chart_json
    except Exception as e:
        # Log the error and return a meaningful response
        print(f"Error generating chart for initiative {initiative}: {str(e)}")
        return {"error": str(e)}, 500

@chairman_blueprint.route("/chairman/executer-chart")
def generate_executer_chart():
    try:
        executer = request.args.get("executer")
        if not executer:
            return {"error": "No executer specified"}, 400

        # Get CSV data
        _, table_data = getCsvData("V3_Data.csv")

        # Convert to DataFrame
        df = pd.DataFrame(table_data)

        # Ensure SDI and Delay_Percent are numeric (convert errors to NaN)
        df['SDI'] = pd.to_numeric(df['SDI'], errors='coerce')
        df['Delay_Percent'] = pd.to_numeric(df['Delay_Percent'], errors='coerce')

        # Drop rows where Status_Final is in the specified list
        excluded_statuses = ["On Track", "On Hold", "Finished", "Closure", "Not Started"]
        df = df[~df["Status_Final"].isin(excluded_statuses)]

        # Filter for projects delayed by the specified executer and SDI > 0
        filtered_df = df[(df["Executer"] == executer) & (df["SDI"] > 0)]

        if filtered_df.empty:
            return {"error": "No valid data available for this executer"}, 404

        # Ensure the SN and Initiative columns exist
        required_columns = ["SN", "Initiative"]
        missing_columns = [col for col in required_columns if col not in filtered_df.columns]
        if missing_columns:
            return {"error": f"Required columns missing from data: {', '.join(missing_columns)}"}, 500

        # Create Unique Project Name using Executer + (SN)
        filtered_df['Unique_Project_Name'] = (
            filtered_df['Initiative'] + ' (' + filtered_df['SN'].astype(str) + ')'
        )

        # Melt the DataFrame with SDI and Delay_Percent
        melted_df = filtered_df.melt(
            id_vars=["Unique_Project_Name"],
            value_vars=["SDI", "Delay_Percent"],
            var_name="Metric",
            value_name="Value"
        )

        # Map Metric names to desired legend labels
        metric_mapping = {
            'SDI': 'Performance Delay - SDI',
            'Delay_Percent': 'Progress Delay - Current'
        }
        melted_df['Metric'] = melted_df['Metric'].map(metric_mapping)

        # Generate the chart
        fig = px.bar(
            melted_df,
            x="Unique_Project_Name",
            y="Value",
            color="Metric",
            text="Value",  # Add value labels on top of bars
            barmode="group",  # Side-by-side bars
            title=f"Delayed Projects by Executer: {executer}",
            labels={
                "Unique_Project_Name": "Initiative (SN)",
                "Value": "Metric Value",
                "Metric": "Type"
            },
        )

        # Customize traces for colors and patterns based on Metric names
        for trace in fig.data:
            if trace.name == "Progress Delay - Current":
                trace.marker.color = "red"  # Solid red color for Progress Delay - Current
            elif trace.name == "Performance Delay - SDI":
                trace.marker.color = "lightgrey"  # Grey color for Performance Delay - SDI
                trace.marker.pattern = dict(shape="x")  # Hatched pattern for Performance Delay - SDI

        # Add the red line at y=10
        fig.add_shape(
            type="line",
            x0=-0.5,  # Extend the line slightly left of the first bar
            x1=len(filtered_df["Unique_Project_Name"]) - 0.5,  # Extend to the last bar
            y0=10,
            y1=10,
            line=dict(color="red", width=2, dash="dash"),
            xref="x",
            yref="y"
        )

        # Add border to the plot area
        fig.add_shape(
            type="rect",
            x0=0, x1=1, y0=0, y1=1,  # Full plot area
            xref="paper",  # Relative to the paper size
            yref="paper",
            line=dict(color="black", width=1)  # Solid black border
        )

        # Update y-axis configuration with gridlines
        fig.update_yaxes(
            type="log",  # Logarithmic scale
            range=[0, 2.4771],  # Corresponds to roughly 0 to 150 in log scale
            tickvals=[1, 2, 4, 6, 8, 10, 25, 50, 75, 100, 150, 200, 250, 300],  # Custom tick values
            title="Delay %",
            automargin=True,  # Automatically adjust margins
            ticklabelposition="outside",  # Ensure labels are placed outside
            showgrid=True,  # Add horizontal gridlines
            gridcolor="lightgrey",  # Gridline color
            gridwidth=1  # Gridline width
        )

        # Update x-axis configuration with gridlines
        fig.update_xaxes(
            tickangle=90,  # Vertical labels
            title="Initiative (SN)",
            automargin=True,  # Automatically adjust margins
            ticklabelposition="outside",  # Ensure labels are placed outside
            showgrid=True,  # Add vertical gridlines
            gridcolor="lightgrey",  # Gridline color
            gridwidth=1  # Gridline width
        )

        # Adjust overall layout to add padding, reposition legend, and center title
        fig.update_layout(
            plot_bgcolor="rgba(0,0,0,0)",  # Transparent background
            paper_bgcolor="rgba(255,255,255,1)",  # White border
            margin=dict(l=60, r=60, t=120, b=60),  # Adjust overall plot margins (increased top margin for legend)
            legend=dict(
                orientation="h",  # Horizontal orientation
                yanchor="bottom",
                y=1.02,  # Position above the plot
                xanchor="center",
                x=0.5  # Centered horizontally
            ),
            title=dict(
                text=f"Delayed Projects by Executer: {executer}",  # Chart title
                x=0.5,  # Center horizontally
                y=0.95,  # Position vertically (adjust as needed)
                xanchor='center',
                yanchor='top',
                font=dict(
                    size=20  # Adjust font size as needed
                )
            )
        )

        # Convert the chart to JSON
        chart_json = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return chart_json
    except Exception as e:
        # Log the error and return a meaningful response
        print(f"Error generating chart for executer {executer}: {str(e)}")
        return {"error": str(e)}, 500


@chairman_blueprint.route("/chairman/default-chart")
def generate_default_chart():
    try:
        # Get CSV data
        _, table_data = getCsvData("V3_Data.csv")

        # Convert to DataFrame
        df = pd.DataFrame(table_data)

        # Ensure SDI is numeric (convert errors to NaN)
        df['SDI'] = pd.to_numeric(df['SDI'], errors='coerce')

        # Drop rows where Status_Final is in the excluded list
        excluded_statuses = ["On Track", "On Hold", "Finished", "Closure", "Not Started"]
        df = df[~df["Status_Final"].isin(excluded_statuses)]

        # Filter delayed projects (SDI > 0)
        delayed_df = df[df["SDI"] > 0]

        # Count delayed projects by Initiative and Executor
        stacked_data = delayed_df.groupby(["Initiative", "Executer"]).size().reset_index(name="Delayed_Count")

        # Calculate totals for each bar
        bar_totals = stacked_data.groupby("Initiative")["Delayed_Count"].sum().reset_index()

        # Generate the stacked bar chart
        fig = px.bar(
            stacked_data,
            x="Initiative",
            y="Delayed_Count",
            color="Executer",  # Stack bars by Executor
            title="Count of Delayed Projects by Initiative (Stacked by Executor)",
            labels={
                "Initiative": "Initiative",
                "Delayed_Count": "Count of Delayed Projects",
                "Executer": "Executor"
            },
            text="Executer",  # Add executor names inside the stacked areas
        )
        fig.update_traces(textposition="inside")  # Position text inside the bars

        # Add totals on top of each bar
        for row in bar_totals.itertuples():
            fig.add_annotation(
                x=row.Initiative,
                y=row.Delayed_Count,
                text=str(row.Delayed_Count),
                showarrow=False,
                font=dict(size=12, color="black"),
                yshift=10,  # Position the text above the bar
            )

        fig.update_layout(
            xaxis_tickangle=90,  # Vertical labels
            xaxis_title="Initiative",
            barmode="stack",  # Stack the bars
            height=600,  # Increase chart height
             title=dict(
                text=f"Count of Delayed Projects By Initiative (Stacked By Executer)",  # Chart title
                x=0.5,  # Center horizontally
                y=0.95,  # Position vertically (adjust as needed)
                xanchor='center',
                yanchor='top',
                font=dict(
                    size=20  # Adjust font size as needed
                )
            )
        )

        # Convert the chart to JSON
        chart_json = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return chart_json
    except Exception as e:
        # Log the error and return a meaningful response
        print(f"Error generating default chart: {str(e)}")
        return {"error": str(e)}, 500


@chairman_blueprint.route("/chairman/default-chart-2")
def generate_executor_stacked_chart():
    try:
        # Get CSV data
        _, table_data = getCsvData("V3_Data.csv")

        # Convert to DataFrame
        df = pd.DataFrame(table_data)

        # Ensure SDI is numeric (convert errors to NaN)
        df['SDI'] = pd.to_numeric(df['SDI'], errors='coerce')

        # Drop rows where Status_Final is in the excluded list
        excluded_statuses = ["On Track", "On Hold", "Finished", "Closure", "Not Started"]
        df = df[~df["Status_Final"].isin(excluded_statuses)]

        # Filter delayed projects (SDI > 0)
        delayed_df = df[df["SDI"] > 0]

        # Count delayed projects by Executor and Initiative
        stacked_data = delayed_df.groupby(["Executer", "Initiative"]).size().reset_index(name="Delayed_Count")

        # Calculate totals for each bar
        bar_totals = stacked_data.groupby("Executer")["Delayed_Count"].sum().reset_index()

        # Generate the stacked bar chart
        fig = px.bar(
            stacked_data,
            x="Executer",
            y="Delayed_Count",
            color="Initiative",  # Stack bars by Initiative
            title="Total Project Delays by Executor (Stacked by Initiative)",
            labels={
                "Executer": "Executor",
                "Delayed_Count": "Count of Delayed Projects",
                "Initiative": "Initiative"
            },
            text="Initiative",  # Add initiative names inside the stacked areas
        )
        fig.update_traces(textposition="inside")  # Position text inside the bars

        # Add totals on top of each bar
        for row in bar_totals.itertuples():
            fig.add_annotation(
                x=row.Executer,
                y=row.Delayed_Count,
                text=str(row.Delayed_Count),
                showarrow=False,
                font=dict(size=12, color="black"),
                yshift=10,  # Position the text above the bar
            )

        fig.update_yaxes(title="Count of Delayed Projects")
        fig.update_layout(
            xaxis_tickangle=90,  # Vertical labels
            xaxis_title="Executor",
            barmode="stack",  # Stack the bars
            height=600,  # Increase chart height
             title=dict(
                text=f"Count of Delayed Projects By Executer (Stacked By Initiative)",  # Chart title
                x=0.5,  # Center horizontally
                y=0.95,  # Position vertically (adjust as needed)
                xanchor='center',
                yanchor='top',
                font=dict(
                    size=20  # Adjust font size as needed
                )
            )
        )

        # Convert the chart to JSON
        chart_json = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return chart_json
    except Exception as e:
        # Log the error and return a meaningful response
        print(f"Error generating executor chart: {str(e)}")
        return {"error": str(e)}, 500

@chairman_blueprint.route("/chairman/table-data")
def get_table_data():
    try:
        _, table_data = getCsvData("V3_Data.csv")  # Load CSV into table_data
        
        # Convert NaN values to None for JSON compatibility
        cleaned_table_data = pd.DataFrame(table_data).replace({pd.NA: None, float('nan'): None}).to_dict(orient="records")
        
        return json.dumps(cleaned_table_data)  # Convert to JSON
    except Exception as e:
        print(f"Error loading table data: {e}")
        return {"error": str(e)}, 500


@chairman_blueprint.route("/chairman")
def home():
    # Get CSV data
    headers, table_data = getCsvData("V3_Data.csv")
    # Classify data and get delay_df
    summary_data, summary_data_meta, delay_df = classifyData(table_data)

    return render_template(
        "chairman.html", 
        headers=headers, 
        table_data=table_data, 
        summary_data=summary_data, 
        summary_data_meta=summary_data_meta, 
    )
